import logging
from functools import cached_property
from typing import Any

from banal import ensure_list
from followthemoney import EntityProxy, model
from ftmq.util import get_name_symbols

from openaleph_search.index.entities import ENTITY_SOURCE, PROXY_INCLUDES
from openaleph_search.index.indexes import entities_read_index
from openaleph_search.index.mapping import Field
from openaleph_search.query.base import Query
from openaleph_search.query.matching import match_query
from openaleph_search.query.more_like_this import more_like_this_query
from openaleph_search.query.util import field_filter_query
from openaleph_search.settings import Settings
from openaleph_search.transform.util import index_name_keys

log = logging.getLogger(__name__)
settings = Settings()


EXCLUDE_SCHEMATA = [
    s.name for s in model.schemata.values() if s.hidden
]  # Page, Mention
EXCLUDE_DEHYDRATE = ["properties"]


class EntitiesQuery(Query):
    TEXT_FIELDS = [
        f"{Field.NAME}^4",
        f"{Field.NAMES}^3",
        f"{Field.NAME_PARTS}^2",
        Field.CONTENT,
        f"{Field.TEXT}^0.8",
    ]
    PREFIX_FIELD = Field.NAME_PARTS
    HIGHLIGHT_FIELD = Field.CONTENT
    SKIP_FILTERS = [Field.SCHEMA, Field.SCHEMATA]
    SOURCE = ENTITY_SOURCE
    SORT_DEFAULT = []

    @cached_property
    def schemata(self) -> list[str]:
        schemata = self.parser.getlist("filter:schema")
        if len(schemata):
            return schemata
        schemata = self.parser.getlist("filter:schemata")
        if not len(schemata):
            schemata = ["Thing"]
        return schemata

    def get_index(self):
        return entities_read_index(schema=self.schemata)

    def get_query(self) -> dict[str, Any]:
        query = self.get_inner_query()
        if settings.query_function_score:
            return self.wrap_query_function_score(query)
        return query

    def get_inner_query(self) -> dict[str, Any]:
        return super().get_query()

    def get_query_string(self) -> dict[str, Any] | None:
        query = super().get_query_string()
        if self.schemata != ["Page"]:
            return query
        if query:
            # special case for Page (children of Pages) queries, where filter
            # syntax would not match (e.g. 'names:"Jane" foo')
            query["query_string"]["default_operator"] = "OR"
            return query

    def get_text_query(self) -> list[dict[str, Any]]:
        query = super().get_text_query()

        # Add optional name_symbols and name_keys matches if synonyms is enabled
        if self.parser.synonyms and self.parser.text:
            schema = model["LegalEntity"]
            # Extract symbols and filter only NAME category symbols
            symbols = get_name_symbols(schema, self.parser.text)
            name_symbols = [str(s) for s in symbols if s.category.name == "NAME"]
            if name_symbols:
                query.append(
                    {"terms": {Field.NAME_SYMBOLS: name_symbols, "boost": 0.5}}
                )

            # Generate name_keys from n-grams of query tokens
            # For a query like "acme corporation corruption", we want to match
            # entities with name_keys like "acmecorporation" (ACME Corporation)
            name_keys = self._get_name_keys_ngrams(schema, self.parser.text)
            if name_keys:
                query.append(
                    {"terms": {Field.NAME_KEYS: list(name_keys), "boost": 0.3}}
                )

        return query

    def _get_name_keys_ngrams(self, schema, text: str) -> set[str]:
        """Generate name_keys from n-grams of query tokens.

        For query "acme corporation corruption", generates:
        - "acmecorporation" (2-gram)
        - "acmecorporationcorruption" (3-gram)
        - "corporationcorruption" (2-gram)

        This allows matching entity names that are subsets of the query.
        We focus on 2-4 token combinations as single tokens are already
        covered by name_parts field.
        """
        from openaleph_search.transform.util import clean_tokenize_name

        tokens = clean_tokenize_name(schema, text)
        if not tokens:
            return set()

        name_keys = set()

        # Generate n-grams of length 2-4 tokens
        for start_idx in range(len(tokens)):
            for length in range(2, min(5, len(tokens) - start_idx + 1)):
                end_idx = start_idx + length
                ngram_tokens = tokens[start_idx:end_idx]

                # Pre-check: ensure combined length is reasonable before calling
                # index_name_keys. This avoids processing very short n-grams that
                # won't pass index_name_keys filter
                estimated_length = sum(len(t) for t in ngram_tokens)
                if estimated_length < 6:
                    continue

                # Join tokens with space and use index_name_keys to get the key
                # This ensures we use the same logic as the indexer
                ngram_text = " ".join(ngram_tokens)
                ngram_keys = index_name_keys(schema, [ngram_text])
                name_keys.update(ngram_keys)

        return name_keys

    def get_negative_filters(self) -> list[dict[str, Any]]:
        # exclude hidden schemata unless we explicitly want them
        filters = super().get_negative_filters()
        exclude_schemata = set(EXCLUDE_SCHEMATA) - set(self.schemata)
        filters.append(field_filter_query("schema", exclude_schemata))
        return filters

    def get_index_weight_functions(self) -> list[dict[str, Any]]:
        """Generate index weight functions based on index bucket settings"""
        functions = []

        # Map bucket names to their boost settings
        bucket_boosts = {
            "pages": settings.index_boost_pages,
            "documents": settings.index_boost_documents,
            "intervals": settings.index_boost_intervals,
            "things": settings.index_boost_things,
        }

        # Create boost functions for each bucket with non-default boost
        for bucket_name, boost_value in bucket_boosts.items():
            if boost_value != 1:  # Only add function if boost differs from default
                functions.append(
                    {
                        "filter": {"wildcard": {"_index": f"*entity-{bucket_name}*"}},
                        "weight": boost_value,
                    }
                )

        return functions

    def wrap_query_function_score(self, query: dict[str, Any]) -> dict[str, Any]:
        # Wrap query in function_score to up-score important entities.
        # (thank you, OpenSanctions/yente :))
        functions = [
            {
                "field_value_factor": {
                    "field": Field.NUM_VALUES,
                    # This is a bit of a jiggle factor. Currently, very
                    # large documents (like Vladimir Putin) have a
                    # num_values of ~200, so get a +10 boost.  The order
                    # is modifier(factor * value)
                    "factor": 0.5,
                    "modifier": "sqrt",
                }
            }
        ]

        # Add index weight functions
        functions.extend(self.get_index_weight_functions())

        return {
            "function_score": {
                "query": query,
                "functions": functions,
                "boost_mode": "sum",
            }
        }

    def get_source(self) -> dict[str, Any]:
        """If the parser gets `dehydrate=true`, don't include properties payload
        in the response. This is used in the search views where no detail data
        is needed"""
        if self.parser.dehydrate:
            return {
                "includes": [k for k in PROXY_INCLUDES if k not in EXCLUDE_DEHYDRATE]
            }
        return super().get_source()


class MatchQuery(EntitiesQuery):
    """Given an entity, find the most similar other entities."""

    def __init__(
        self,
        parser,
        entity: EntityProxy | None = None,
        exclude=None,
        datasets=None,
        collection_ids=None,
    ):
        self.entity = entity
        self.exclude = ensure_list(exclude)
        self.datasets = datasets
        self.collection_ids = collection_ids
        super(MatchQuery, self).__init__(parser)

    def get_index(self):
        # Attempt to find only matches within the "matchable" set of entity
        # schemata. In practice this should always return the "things" index.
        schemata = list(self.entity.schema.matchable_schemata)
        return entities_read_index(schema=schemata)

    def get_inner_query(self) -> dict[str, Any]:
        query = match_query(
            self.entity,
            datasets=self.datasets,
            collection_ids=self.collection_ids,
            query=super().get_inner_query(),
        )
        if len(self.exclude):
            exclude = {"ids": {"values": self.exclude}}
            query["bool"]["must_not"].append(exclude)
        return query


class MoreLikeThisQuery(EntitiesQuery):
    """Given an entity, find similar documents/pages based on text content using
    elasticsearch more_like_this query."""

    def __init__(
        self,
        parser,
        entity: EntityProxy | None = None,
        exclude=None,
        datasets=None,
        collection_ids=None,
    ):
        self.entity = entity
        self.exclude = ensure_list(exclude)
        self.datasets = datasets
        self.collection_ids = collection_ids
        super(MoreLikeThisQuery, self).__init__(parser)

    def get_index(self):
        # Target only documents and pages buckets for more_like_this queries
        return entities_read_index(schema="Document")

    def get_inner_query(self) -> dict[str, Any]:
        if not self.entity:
            return {"match_none": {}}

        # Get base query with auth filters from parent class
        base_query = super().get_inner_query()

        # Apply more_like_this query
        mlt_query = more_like_this_query(
            self.entity,
            datasets=self.datasets,
            collection_ids=self.collection_ids,
            parser=self.parser,
            query=base_query,
        )

        if len(self.exclude):
            exclude = {"ids": {"values": self.exclude}}
            mlt_query["bool"]["must_not"].append(exclude)
        return mlt_query


class GeoDistanceQuery(EntitiesQuery):
    """Given an Address entity, find the nearby Address entities via the
    geo_point field"""

    def __init__(self, parser, entity=None, exclude=None, datasets=None):
        self.entity = entity
        self.exclude = ensure_list(exclude)
        self.datasets = datasets
        super().__init__(parser)

    def is_valid(self) -> bool:
        return (
            self.entity is not None
            and self.entity.first("latitude") is not None
            and self.entity.first("longitude") is not None
        )

    def get_query(self):
        if not self.is_valid():
            return {"match_none": {}}
        query = super(GeoDistanceQuery, self).get_query()
        exclude = {"ids": {"values": self.exclude + [self.entity.id]}}
        query["bool"]["must_not"].append(exclude)
        query["bool"]["must"].append({"exists": {"field": "geo_point"}})
        return query

    def get_sort(self):
        """Always sort by calculated distance"""
        if not self.is_valid():
            return []
        return [
            {
                "_geo_distance": {
                    "geo_point": {
                        "lat": self.entity.first("latitude"),
                        "lon": self.entity.first("longitude"),
                    },
                    "order": "asc",
                    "unit": "km",
                    "mode": "min",
                    "distance_type": "plane",  # faster
                }
            }
        ]



from typing import Annotated, Optional

import typer
from anystore.cli import ErrorHandler
from anystore.io import (
    logged_items,
    smart_read,
    smart_stream_json,
    smart_write,
    smart_write_json,
)
from anystore.logging import configure_logging, get_logger
from anystore.util import Took, dump_json
from ftmq.io import smart_read_proxies
from rich import print

from openaleph_search.index import admin, entities, export
from openaleph_search.index.indexer import bulk_actions
from openaleph_search.search.logic import (
    analyze_text,
    make_parser,
    search_body,
    search_query_string,
)
from openaleph_search.settings import Settings, __version__
from openaleph_search.transform.entity import format_parallel

settings = Settings()

cli = typer.Typer(no_args_is_help=True, pretty_exceptions_enable=settings.debug)
cli_search = typer.Typer(name="search", no_args_is_help=True)
cli.add_typer(cli_search, short_help="Execute search queries")
log = get_logger(__name__)

OPT_INPUT_URI = typer.Option("-", "-i", help="Input uri, default stdin")
OPT_OUTPUT_URI = typer.Option("-", "-o", help="Output uri, default stdout")
OPT_DATASET = typer.Option(..., "-d", help="Dataset")

OPT_SEARCH_ARGS = Annotated[
    Optional[str],
    typer.Option(
        help="Query parser args and filters (e.g. `filter:dataset=my_dataset`)"
    ),
]
OPT_SEARCH_FORMAT = Annotated[
    Optional[str], typer.Option(help="Output format (raw, parsed)")
]


@cli.callback(invoke_without_command=True)
def cli_openaleph_search(
    version: Annotated[Optional[bool], typer.Option(..., help="Show version")] = False,
    settings: Annotated[
        Optional[bool], typer.Option(..., help="Show current settings")
    ] = False,
):
    if version:
        print(__version__)
        raise typer.Exit()
    if settings:
        settings_ = Settings()
        print(settings_)
        raise typer.Exit()
    configure_logging()


@cli.command("upgrade")
def cli_upgrade():
    """Upgrade the index mappings or create if they don't exist"""
    with ErrorHandler(log):
        admin.upgrade_search()


@cli.command("reset")
def cli_reset():
    """Drop all data and indexes and re-upgrade"""
    with ErrorHandler(log):
        admin.delete_index()
        admin.upgrade_search()


@cli.command("format-entities")
def cli_format_entities(
    input_uri: str = OPT_INPUT_URI,
    output_uri: str = OPT_OUTPUT_URI,
    dataset: str = OPT_DATASET,
):
    """Transform entities into index actions"""
    with ErrorHandler(log):
        entities = smart_read_proxies(input_uri)
        formatted = logged_items(
            format_parallel(dataset, entities), "Format", 10_000, "Entity", log
        )
        smart_write_json(output_uri, formatted)


@cli.command("index-entities")
def cli_index_entities(
    input_uri: str = OPT_INPUT_URI,
    dataset: str = OPT_DATASET,
):
    """Index entities into given dataset"""
    with ErrorHandler(log):
        entities.index_bulk(dataset, smart_read_proxies(input_uri))


@cli.command("index-actions")
def cli_index_actions(input_uri: str = OPT_INPUT_URI):
    """Index a stream of actions"""
    with ErrorHandler(log), Took() as t:
        actions = smart_stream_json(input_uri)
        bulk_actions(actions)
        log.info("Index actions complete.", input_uri=input_uri, took=t.took)


@cli.command("dump-actions")
def cli_dump_actions(
    output_uri: str = OPT_OUTPUT_URI,
    index: str | None = None,
    args: OPT_SEARCH_ARGS = None,
):
    """Export index documents (Actions) by given criteria. For entity indexes,
    this DOESN'T include all necessary data to re-index!"""
    with ErrorHandler(log):
        parser = make_parser(args=args)
        actions = logged_items(
            export.export_index_actions(index, parser), "Export", 10_000, "Action", log
        )
        smart_write_json(output_uri, actions)


@cli_search.command("query-string")
def cli_search_query(
    q: str,
    args: OPT_SEARCH_ARGS = None,
    output_uri: str = OPT_OUTPUT_URI,
    output_format: OPT_SEARCH_FORMAT = "raw",
):
    """Search using elastic 'query_string' using the `EntitiesQuery` class"""
    res = search_query_string(q, args)
    data = dump_json(dict(res), clean=True, newline=True)
    smart_write(output_uri, data)


@cli_search.command("body")
def cli_search_body(
    input_uri: str = OPT_INPUT_URI,
    output_uri: str = OPT_OUTPUT_URI,
    output_format: OPT_SEARCH_FORMAT = "raw",
    index: str | None = None,
):
    """Search with raw json body for query"""
    body = smart_read(input_uri, serialization_mode="json")
    res = search_body(body, index)
    data = dump_json(dict(res), clean=True, newline=True)
    smart_write(output_uri, data)


@cli.command("analyze")
def cli_analyze(
    input_uri: str = OPT_INPUT_URI,
    field: Annotated[
        str,
        typer.Option(
            help="Field to analyze with (e.g. 'content', 'text', 'properties.bodyText')"
        ),
    ] = "content",
    schema: Annotated[
        str, typer.Option(help="Schema to use for field analysis")
    ] = "LegalEntity",
    tokens_only: Annotated[
        bool, typer.Option(help="Return only unique token strings")
    ] = False,
    output_uri: str = OPT_OUTPUT_URI,
):
    """Analyze text using Elasticsearch analyzers from field mappings"""
    with ErrorHandler(log):
        text = smart_read(input_uri, mode="r")
        res = analyze_text(text, field=field, schema=schema, tokens_only=tokens_only)
        if tokens_only:
            data = "\n".join(sorted(res)) + "\n"
        else:
            data = dump_json(dict(res), clean=True, newline=True)
        smart_write(output_uri, data)



from typing import Generator

from anystore.logging import get_logger
from elasticsearch.helpers import scan

from openaleph_search.core import get_es
from openaleph_search.index.indexer import Action
from openaleph_search.parse.parser import SearchQueryParser
from openaleph_search.query.base import Query
from openaleph_search.settings import Settings

log = get_logger(__name__)


def export_index_actions(
    index: str | None = None,
    parser: SearchQueryParser | None = None,
    include_excluded_fields: bool = False,
) -> Generator[Action, None, None]:
    """Export all documents from an index as Action objects.

    Note: By default, this exports only fields stored in _source. For entities,
    several fields are excluded from _source for storage optimization:
    - name_keys, name_parts, name_symbols, name_phonetic (retrievable via docvalues)
    - content (retrievable via stored_fields if the entity schema was
        `Document`). If the entity schema is not a Document and has had a value in
        `indexText` during indexing, this data is NOT retrievable (and therefore not
        recoverable) via this export and can't be used to fully re-index from this
        exported data!
    - text (NOT retrievable - excluded and not stored, but populated from other
        properties via copy_to, so can be regenerated)
    - name (NOT retrievable - excluded and not stored, but populated from other
        properties via copy_to, so can be regenerated)
    - property groups (NOT exported - populated via copy_to, can be regenerated)

    Given this limitation about `text` and `content` fields, this function CAN'T
    be used to back up and restore an entity index reliable.

    Args:
        index: Index name, pattern, or prefix to export from (e.g., "my-index",
            "my-index-*"), defaults to index prefix from settings
        parser: Optional SearchQueryParser to filter documents (defaults to match_all)
        include_excluded_fields: If True, retrieve fields excluded from _source:
            - Keyword fields via docvalues (name_keys, name_parts, etc.)
            - Full-text content via stored_fields
            - Note: 'text' field is NOT retrievable

    Yields:
        Action objects for each document in the index
    """
    from openaleph_search.index.mapping import Field

    es = get_es()
    settings = Settings()
    index = index or f"{settings.index_prefix}-*"

    if parser is None:
        query = {"match_all": {}}
    else:
        query = Query(parser).get_query()

    log.info("Starting index export: %s" % index)

    # Build query body
    query_body = {"query": query}

    # If we need excluded fields, request them via docvalues and stored_fields
    # - Keyword fields (name_keys, etc.) use docvalues
    # - Full-text content field uses stored_fields (has store: true)
    # - text field is NOT retrievable (excluded from _source, not stored)
    # - property groups (addresses, dates, etc.) are NOT exported (populated via copy_to)
    if include_excluded_fields:
        docvalue_fields = [
            Field.NAME_KEYS,
            Field.NAME_PARTS,
            Field.NAME_SYMBOLS,
            Field.NAME_PHONETIC,
        ]
        query_body["docvalue_fields"] = docvalue_fields
        query_body["stored_fields"] = [Field.CONTENT]
        query_body["_source"] = True

    for hit in scan(es, index=index, query=query_body, preserve_order=False):
        action: Action = {
            "_id": hit["_id"],
            "_index": hit["_index"],
            "_source": hit["_source"],
        }

        # Merge excluded fields back into _source if requested
        if include_excluded_fields and "fields" in hit:
            for field_name, field_values in hit["fields"].items():
                # Docvalues always return arrays
                action["_source"][field_name] = field_values

        yield action



from typing import Any

from openaleph_search.index.mapping import Field
from openaleph_search.settings import Settings

settings = Settings()


def get_highlighter(
    field: str, query: dict[str, Any] | None = None, count: int | None = None
) -> dict[str, Any]:
    # Content field - configurable highlighting
    if field == Field.CONTENT:
        if settings.highlighter_fvh_enabled:
            # FVH (Fast Vector Highlighter) configuration
            highlighter = {
                "type": "fvh",
                "fragment_size": settings.highlighter_fragment_size,
                # "fragment_offset": 50,
                "number_of_fragments": count
                or settings.highlighter_number_of_fragments,
                "phrase_limit": settings.highlighter_phrase_limit,  # lower than default (256) for better memory performance  # noqa: B950
                "order": "score",  # Best fragments first
                "boundary_scanner": "chars",  # FVH needs 'chars'
                "boundary_max_scan": settings.highlighter_boundary_max_scan,  # better sentence detection  # noqa: B950
                # Explicit boundary chars added for csv/json/html/code raw text
                "boundary_chars": '.\t\n ,!?;_-=(){}[]<>|"',
                "no_match_size": settings.highlighter_no_match_size,  # Hard limit when no boundary/match found  # noqa: B950
                "fragmenter": "span",  # More precise fragment boundaries
                # "pre_tags": ["<em class='highlight-content'>"],
                # "post_tags": ["</em>"],
                "max_analyzed_offset": settings.highlighter_max_analyzed_offset,  # Handle large documents  # noqa: B950
            }
        else:
            # Unified highlighter with sentence boundary scanner
            highlighter = {
                "type": "unified",
                "fragment_size": settings.highlighter_fragment_size,
                "number_of_fragments": count
                or settings.highlighter_number_of_fragments,
                "order": "score",
                "boundary_scanner": "sentence",  # Use sentence boundary scanner
                "no_match_size": settings.highlighter_no_match_size,
                # "pre_tags": ["<em class='highlight-content'>"],
                # "post_tags": ["</em>"],
                "max_analyzed_offset": settings.highlighter_max_analyzed_offset,
            }
        if query:
            highlighter["highlight_query"] = query
        return highlighter
    # Human-readable names - exact highlighting
    if field == Field.NAME:
        highlighter = {
            "type": "unified",  # Good for mixed content
            "fragment_size": 200,  # Longer to capture full names/titles
            "number_of_fragments": 3,
            "fragmenter": "simple",  # Don't break names awkwardly
            "pre_tags": [""],  # No markup
            "post_tags": [""],  # No markup
            # "pre_tags": ["<em class='highlight-name'>"],
            # "post_tags": ["</em>"],
        }
        return highlighter
    # Keyword names - simple exact matching
    if field == Field.NAMES:
        return {
            "type": "plain",  # Fast for keyword fields
            "number_of_fragments": 3,
            "max_analyzed_offset": 999999,  # probably many names
            "pre_tags": [""],  # No markup
            "post_tags": [""],  # No markup
        }
    # other fields - leftovers, minimal highlighting if possible (not important)
    plain = {
        "type": "plain",  # Fastest option
        "fragment_size": 150,  # Shorter since less important
        "number_of_fragments": 1,  # Just one fragment
        # "max_analyzed_offset": 999999,  # Handle large documents
        # "pre_tags": ["<em class='highlight-text'>"],
        # "post_tags": ["</em>"],
    }
    if query:
        plain["highlight_query"] = query
    return plain



import logging

from followthemoney import EntityProxy

from openaleph_search.index.mapping import Field
from openaleph_search.query.util import BoolQuery, bool_query, none_query

log = logging.getLogger(__name__)


def more_like_this_query(
    entity: EntityProxy,
    datasets: list[str] | None = None,
    collection_ids: list[str] | None = None,
    query: BoolQuery | None = None,
    parser=None,
) -> BoolQuery:
    """Given an entity, build a more_like_this query that will find
    similar documents/pages based on text content using the entity's
    document ID in elasticsearch."""

    if not entity.id:
        return none_query()

    if query is None:
        query = bool_query()

    # Don't match the query entity itself
    must_not = []
    if entity.id is not None:
        must_not.append({"ids": {"values": [entity.id]}})
    if len(must_not):
        query["bool"]["must_not"].extend(must_not)

    # Apply dataset/collection filters
    if collection_ids:
        query["bool"]["filter"].append({"terms": {"collection_id": collection_ids}})
    elif datasets:
        query["bool"]["filter"].append({"terms": {"dataset": datasets}})

    # Get configurable parameters from parser, with sensible defaults
    min_doc_freq = 2
    minimum_should_match = "20%"
    min_term_freq = 2
    max_query_terms = 50

    if parser is not None:
        min_doc_freq = parser.get_mlt_min_doc_freq()
        minimum_should_match = parser.get_mlt_minimum_should_match()
        min_term_freq = parser.get_mlt_min_term_freq()
        max_query_terms = parser.get_mlt_max_query_terms()

    # Build the more_like_this query using document ID
    mlt_query = {
        "more_like_this": {
            "fields": [Field.CONTENT, Field.TEXT, Field.NAME],
            "like": [{"_id": entity.id}],
            "min_term_freq": min_term_freq,
            "max_query_terms": max_query_terms,
            "min_doc_freq": min_doc_freq,
            "minimum_should_match": minimum_should_match,
        }
    }

    # Add the more_like_this query to the main query
    query["bool"]["must"].append(mlt_query)

    # exclude Page entities
    query["bool"]["must_not"].append({"term": {"schema": "Page"}})

    return query



from typing import TypeAlias

from anystore.functools import weakref_cache
from followthemoney import Schema, model
from followthemoney.dataset.util import dataset_name_check

SchemaType: TypeAlias = Schema | str


@weakref_cache
def valid_dataset(dataset: str) -> str:
    return dataset_name_check(dataset)


@weakref_cache
def ensure_schema(schema: SchemaType) -> Schema:
    schema_ = model.get(schema)
    if schema_ is not None:
        return schema_
    raise ValueError(f"Invalid schema: `{schema}`")



from anystore.settings import BaseSettings
from pydantic import AliasChoices, Field, HttpUrl
from pydantic_settings import SettingsConfigDict

__version__ = "5.1.1"

MAX_PAGE = 9999
BULK_PAGE = 1000


class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        env_file=".env", env_prefix="openaleph_search_", extra="ignore"
    )

    testing: bool = Field(
        default=False, validation_alias=AliasChoices("testing", "debug")
    )

    auth: bool = False
    """Set to true when using with OpenAleph"""

    auth_field: str = "dataset"
    """Default field to filter/apply auth on"""

    uri: HttpUrl | list[HttpUrl] = Field(
        default=HttpUrl("http://localhost:9200"), alias="openaleph_elasticsearch_uri"
    )

    ingest_uri: HttpUrl | list[HttpUrl] | None = Field(
        default=None, alias="openaleph_elasticsearch_ingest_uri"
    )
    """Optional dedicated URI(s) for ingest (pre-index) operations (nodes with
    dedicated ingest role that might do some enrichment). Falls back to uri if
    not set."""

    timeout: int = 60
    max_retries: int = 3
    retry_on_timeout: bool = True

    # Connection pool settings for AsyncElasticsearch
    connection_pool_limit_per_host: int = 25

    indexer_concurrency: int = 8
    indexer_chunk_size: int = 1000
    indexer_max_chunk_bytes: int = 5 * 1024 * 1024  # 5mb

    index_shards: int = 10
    index_replicas: int = 0
    index_prefix: str = "openaleph"
    index_write: str = "v1"
    index_read: str | list[str] = ["v1"]
    index_expand_clause_limit: int = 10
    index_delete_by_query_batchsize: int = 100
    index_namespace_ids: bool = True
    index_refresh_interval: str = "1s"

    # configure different weights for indices
    index_boost_intervals: int = 1
    index_boost_things: int = 1
    index_boost_documents: int = 1
    index_boost_pages: int = 1

    # enable/disable function_score wrapper for performance tuning
    query_function_score: bool = False

    # enable/disable term vectors and offsets for content field (used for highlighting)
    content_term_vectors: bool = True

    # Highlighter configuration
    highlighter_fvh_enabled: bool = True
    highlighter_fragment_size: int = 200
    highlighter_number_of_fragments: int = 3
    highlighter_phrase_limit: int = 64
    highlighter_boundary_max_scan: int = 100
    highlighter_no_match_size: int = 300
    highlighter_max_analyzed_offset: int = 999999



"""High level search interface"""

from typing import Any
from urllib.parse import parse_qsl

from elastic_transport import ObjectApiResponse

from openaleph_search.core import get_es
from openaleph_search.index.indexes import entities_read_index, schema_index
from openaleph_search.parse.parser import SearchQueryParser
from openaleph_search.query.queries import EntitiesQuery


def make_parser(q: str | None = None, args: str | None = None) -> SearchQueryParser:
    """Build a parser object from arbitrary query string and url args"""
    if args:
        _args = parse_qsl(args, keep_blank_values=True)
    else:
        _args = []
    if q:
        _args.insert(0, ("q", q))
    return SearchQueryParser(_args)


def search_query_string(q: str, args: str | None = None) -> ObjectApiResponse:
    """Search using `query_string` with optional parser args"""
    _args = parse_qsl(args, keep_blank_values=True)
    if "q" in dict(_args):
        raise RuntimeError("Invalid query, must not contain `q` in args")
    _args.insert(0, ("q", q))
    parser = SearchQueryParser(_args)
    query = EntitiesQuery(parser)
    return query.search()


def search_body(body: dict[str, Any], index: str | None = None) -> ObjectApiResponse:
    es = get_es()
    index = index or entities_read_index()
    return es.search(index=index, body=body)


def analyze_text(
    text: str,
    field: str,
    schema: str = "LegalEntity",
    tokens_only: bool = False,
) -> ObjectApiResponse | set[str]:
    """Analyze text using field mapping for given schema.

    Args:
        text: The text to analyze
        field: Field name to use for analysis
            (e.g., "content", "text", "properties.bodyText")
        schema: Schema to use for field-based analysis
            (default: "LegalEntity")
        tokens_only: If True, return only unique token strings instead of
            full response (default: False)

    Returns:
        ObjectApiResponse with tokens information, or set of unique token
        strings if tokens_only=True
    """
    es = get_es()
    index = schema_index(schema, "v1")
    res = es.indices.analyze(index=index, field=field, text=text)

    if tokens_only:
        return {token["token"] for token in res["tokens"]}
    return res



import contextlib

from anystore.logging import get_logger
from anystore.types import SDict
from banal import ensure_list

from openaleph_search.core import get_es
from openaleph_search.index.mapping import ANALYZE_SETTINGS  # , Field
from openaleph_search.settings import Settings

log = get_logger(__name__)
settings = Settings()


def index_name(name: str, version: str) -> str:
    return "-".join((settings.index_prefix, name, version))


def check_response(index, res):
    """Check if a request succeeded."""
    if res.get("status", 0) > 399 and not res.get("acknowledged"):
        error = res.get("error", {}).get("reason")
        log.error("Index [%s] error: %s" % (index, error))
        return False
    return True


def refresh_sync(sync: bool | None = False) -> bool:
    return settings.testing or bool(sync)


def unpack_result(res: SDict) -> SDict | None:
    """Turn a document hit from ES into a more traditional JSON object."""
    error = res.get("error")
    if error is not None:
        raise RuntimeError("Query error: %r" % error)
    if res.get("found") is False:
        return
    data = res.get("_source", {})
    data["id"] = res.get("_id")
    data["_index"] = res.get("_index")

    _score = res.get("_score")
    if _score is not None and _score != 0.0 and "score" not in data:
        data["score"] = _score

    if "highlight" in res:
        data["highlight"] = res["highlight"]
        # data["highlight"] = []
        # for value in res.get("highlight", {}).values():
        #     data["highlight"].extend(value)

    data["_sort"] = ensure_list(res.get("sort"))
    return data


def check_settings_changed(updated, existing):
    """Since updating the settings requires closing the index, we don't
    want to do it unless it's really needed. This will check if all the
    updated settings are already in effect."""
    if not isinstance(updated, dict) or not isinstance(existing, dict):
        return updated != existing
    for key, value in list(updated.items()):
        if check_settings_changed(value, existing.get(key)):
            return True
    return False


def index_settings(
    shards: int | None = settings.index_shards,
    replicas: int | None = settings.index_replicas,
):
    """Configure an index in ES with support for text transliteration."""
    if settings.testing:
        shards = 1
        replicas = 0
    return {
        **ANALYZE_SETTINGS,
        "index": {
            "number_of_shards": str(shards),
            "number_of_replicas": str(replicas),
            "refresh_interval": (
                "10ms" if settings.testing else settings.index_refresh_interval
            ),
            "similarity": {
                # We use this for names, to avoid over-penalizing entities with many names.
                "weak_length_norm": {
                    # BM25 is the default similarity algorithm.
                    "type": "BM25",
                    # 0.75 is the default
                    "b": 0.25,
                }
            },
            # storage optimization to put similar docs together for better compression:
            # https://www.elastic.co/docs/reference/elasticsearch/index-settings/sorting
            # "sort": {
            #     "field": [Field.SCHEMA, Field.DATASET]
            # }
        },
    }


@contextlib.contextmanager
def bulk_indexing_mode(refresh_interval: str = "600s"):
    """Context manager for bulk indexing with optimized settings.

    Phase 1: Sets optimized settings for maximum bulk indexing performance:
    - Async translog for better write performance
    - Longer sync and refresh intervals
    - No replicas during bulk load

    Phase 2: Restores normal production settings after bulk operations complete:
    - Request-level translog durability
    - Normal refresh interval
    - Configured number of replicas

    Args:
        refresh_interval: The refresh interval to use during bulk operations (default: "600s")

    Example:
        with bulk_indexing_mode():
            # Perform bulk operations with optimized settings
            bulk_index_entities(large_entity_batch)
        # Settings are automatically restored to normal production values
    """
    es = get_es()
    index_pattern = f"{settings.index_prefix}-entity-*"

    log.info("Entering bulk indexing mode with optimized settings")

    # Phase 1: Set bulk indexing settings
    bulk_settings = {
        "index": {
            "translog.durability": "async",
            "translog.sync_interval": "60s",
            "refresh_interval": refresh_interval,
            "number_of_replicas": 0,
        }
    }

    try:
        res = es.indices.put_settings(index=index_pattern, body=bulk_settings)
        if not check_response(index_pattern, res):
            raise RuntimeError("Failed to set bulk indexing settings")

        log.info(f"Set bulk indexing settings for entity indices: {index_pattern}")
        yield

    except Exception as e:
        log.error(
            f"Failed to set bulk indexing settings for entity indices {index_pattern}: {e}"
        )
        raise
    finally:
        # Phase 2: Restore normal settings
        log.info("Exiting bulk indexing mode, restoring normal settings")

        normal_settings = {
            "index": {
                "translog.durability": "request",
                "refresh_interval": settings.index_refresh_interval,
                "number_of_replicas": settings.index_replicas,
            }
        }

        try:
            res = es.indices.put_settings(index=index_pattern, body=normal_settings)
            if check_response(index_pattern, res):
                log.info(
                    f"Restored normal settings for entity indices: {index_pattern}"
                )
            else:
                log.error("Failed to restore normal settings after bulk indexing")
        except Exception as e:
            log.error(
                f"Failed to restore normal settings for entity indices {index_pattern}: {e}"
            )



"""
Index mappings
"""

from collections import defaultdict as ddict
from typing import Any, Iterable, TypeAlias

from followthemoney import model
from followthemoney.types import registry

from openaleph_search.settings import Settings
from openaleph_search.util import SchemaType

settings = Settings()

MappingProperty: TypeAlias = dict[str, list[str] | str]
Mapping: TypeAlias = dict[str, MappingProperty]

# MAPPING SHORTCUTS #
DEFAULT_ANALYZER = "default"
DEFAULT_NORMALIZER = "default"
ICU_ANALYZER = "icu-default"
ICU_NORMALIZER = "icu-default"
HTML_ANALYZER = "strip-html"
KW_NORMALIZER = "kw-normalizer"
NAME_KW_NORMALIZER = "name-kw-normalizer"
DATE_FORMAT = "yyyy-MM-dd'T'HH||yyyy-MM-dd'T'HH:mm||yyyy-MM-dd'T'HH:mm:ss||yyyy-MM-dd||yyyy-MM||yyyy||strict_date_optional_time"  # noqa: B950
NUMERIC_TYPES = (registry.number, registry.date)

# INDEX SETTINGS #
ANALYZE_SETTINGS = {
    "analysis": {
        "char_filter": {
            "remove_punctuation": {
                "type": "pattern_replace",
                "pattern": "[^\\p{L}\\p{N}]",
                "replacement": " ",
            },
            "squash_spaces": {
                "type": "pattern_replace",
                "pattern": "\\s+",
                "replacement": " ",
            },
        },
        "normalizer": {
            ICU_NORMALIZER: {
                "type": "custom",
                "filter": ["icu_folding"],
            },
            NAME_KW_NORMALIZER: {
                "type": "custom",
                "char_filter": ["remove_punctuation", "squash_spaces"],
                "filter": ["lowercase", "asciifolding", "trim"],
            },
            KW_NORMALIZER: {
                "type": "custom",
                "filter": ["trim"],
            },
        },
        "analyzer": {
            ICU_ANALYZER: {
                "char_filter": ["html_strip"],
                "tokenizer": "icu_tokenizer",
                "filter": [
                    "icu_folding",
                    "icu_normalizer",
                ],
            },
            HTML_ANALYZER: {
                "tokenizer": "standard",
                "char_filter": ["html_strip"],
                "filter": ["lowercase", "asciifolding", "trim"],
            },
        },
    },
}


# FIELD NAMES #
class Field:
    DATASET = "dataset"
    DATASETS = "datasets"
    SCHEMA = "schema"
    SCHEMATA = "schemata"
    CAPTION = "caption"
    NAME = "name"
    NAMES = "names"
    NAME_KEYS = "name_keys"
    NAME_PARTS = "name_parts"
    NAME_SYMBOLS = "name_symbols"
    NAME_PHONETIC = "name_phonetic"
    PROPERTIES = "properties"
    NUMERIC = "numeric"
    GEO_POINT = "geo_point"
    CONTENT = "content"
    TEXT = "text"
    TAGS = "tags"

    NUMERIC = "numeric"
    PROPERTIES = "properties"

    CREATED_AT = "created_at"
    UPDATED_AT = "updated_at"

    # align with nomenklatura
    FIRST_SEEN = "first_seen"
    LAST_SEEN = "last_seen"
    LAST_CHANGE = "last_change"
    REFERENTS = "referents"

    # leaked from OpenAleph app, probably deprecated in v6
    ROLE = "role_id"
    PROFILE = "profile_id"
    ORIGIN = "origin"
    COLLECTION_ID = "collection_id"
    MUTABLE = "mutable"

    # length norm
    NUM_VALUES = "num_values"

    # index metadata
    INDEX_BUCKET = "index_bucket"
    INDEX_VERSION = "index_version"
    INDEX_TS = "indexed_at"


FULLTEXTS = [Field.CONTENT, Field.TEXT]


# FIELD TYPES #
class FieldType:
    DATE = {"type": "date"}
    PARTIAL_DATE = {"type": "date", "format": DATE_FORMAT}
    # actual text content (bodyText et. al), optimized for highlighting and
    # termvectors
    CONTENT = {
        "type": "text",
        "analyzer": ICU_ANALYZER,
        "search_analyzer": ICU_ANALYZER,
        "index_phrases": True,  # shingles
        "term_vector": (
            "with_positions_offsets" if settings.content_term_vectors else False
        ),
    }
    # additional text copied over from other properties for arbitrary lookups
    TEXT = {"type": "text", "analyzer": HTML_ANALYZER, "search_analyzer": HTML_ANALYZER}

    KEYWORD = {"type": "keyword", "normalizer": KW_NORMALIZER}
    KEYWORD_COPY = {"type": "keyword", "copy_to": Field.TEXT}
    NUMERIC = {"type": "double"}
    INTEGER = {"type": "integer"}
    GEOPOINT = {"type": "geo_point"}
    BOOL = {"type": "boolean"}

    # No length normalization for names. Merged entities have a lot of names,
    # and we don't want to penalize them for that.
    NAME = {"type": "text", "similarity": "weak_length_norm", "store": True}

    # custom normalized name keywords (used for term aggregations et. al)
    # this is used for registry.name.group. store for nicer highlighting
    NAME_KEYWORD = {
        "type": "keyword",
        "normalizer": NAME_KW_NORMALIZER,
        "store": True,
    }


TYPE_MAPPINGS = {
    registry.text: {"type": "text", "index": False},
    registry.html: {"type": "text", "index": False},
    registry.json: {"type": "text", "index": False},
    registry.date: FieldType.PARTIAL_DATE,
}

GROUPS = {t.group for t in registry.groups.values() if t.group}


# These fields will be pruned from the _source field after the document has been
# indexed, but before the _source field is stored. We can still search on these
# fields, even though they are not in the stored and returned _source.
SOURCE_EXCLUDES = list(
    sorted(
        [
            *GROUPS,
            Field.TEXT,
            Field.CONTENT,
            Field.NAME,
            Field.NAME_KEYS,
            Field.NAME_PARTS,
            Field.NAME_SYMBOLS,
            Field.NAME_PHONETIC,
        ]
    )
)


# base property mapping without specific schema fields
BASE_MAPPING = {
    Field.DATASET: FieldType.KEYWORD,
    Field.SCHEMA: FieldType.KEYWORD,
    Field.SCHEMATA: FieldType.KEYWORD,
    # for fast label display
    Field.CAPTION: FieldType.KEYWORD,
    # original names as matching (text) field
    Field.NAME: FieldType.NAME,
    # names keywords, a bit normalized
    Field.NAMES: FieldType.NAME_KEYWORD,
    # name normalizations for filters and matching
    Field.NAME_KEYS: FieldType.KEYWORD,
    Field.NAME_PARTS: FieldType.KEYWORD_COPY,
    Field.NAME_SYMBOLS: FieldType.KEYWORD,
    Field.NAME_PHONETIC: FieldType.KEYWORD,
    # all entities can reference geo points
    Field.GEO_POINT: FieldType.GEOPOINT,
    # references to other entities (after merging)
    Field.REFERENTS: FieldType.KEYWORD,
    # full text
    Field.CONTENT: FieldType.CONTENT,
    Field.TEXT: FieldType.TEXT,
    # tagging
    Field.TAGS: FieldType.KEYWORD,
    # processing metadata
    Field.UPDATED_AT: FieldType.DATE,
    Field.CREATED_AT: FieldType.DATE,
    # data metadata, provenance
    Field.LAST_CHANGE: FieldType.DATE,
    Field.LAST_SEEN: FieldType.DATE,
    Field.FIRST_SEEN: FieldType.DATE,
    Field.ORIGIN: FieldType.KEYWORD,
    # OpenAleph leaked context data probably deprecated soon
    Field.ROLE: FieldType.KEYWORD,
    Field.PROFILE: FieldType.KEYWORD,
    Field.COLLECTION_ID: FieldType.KEYWORD,
    Field.MUTABLE: FieldType.BOOL,
    # length normalization
    Field.NUM_VALUES: FieldType.INTEGER,
    # index metadata
    Field.INDEX_BUCKET: {**FieldType.KEYWORD, "index": False},
    Field.INDEX_VERSION: {**FieldType.KEYWORD, "index": False},
    Field.INDEX_TS: {
        **FieldType.DATE,
        "index": True,
    },  # we might want to filter on this
}

# combined fields for emails, countries, ...
GROUP_MAPPING = {
    group: TYPE_MAPPINGS.get(type_, FieldType.KEYWORD)
    for group, type_ in registry.groups.items()
    if group not in BASE_MAPPING
}

# used for efficient sorting
NUMERIC_MAPPING = {
    **{
        prop.name: FieldType.NUMERIC
        for prop in model.properties
        if prop.type in NUMERIC_TYPES
    },
    **{
        group: FieldType.NUMERIC
        for group, type_ in registry.groups.items()
        if type_ in NUMERIC_TYPES
    },
}


def property_field_name(prop: str) -> str:
    return f"{Field.PROPERTIES}.{prop}"


def make_object_type(properties: dict[str, MappingProperty]) -> dict[str, Any]:
    return {"type": "object", "properties": properties}


def make_mapping(properties: Mapping) -> dict[str, Any]:
    return {
        "date_detection": False,
        "dynamic": False,
        "_source": {"excludes": SOURCE_EXCLUDES},
        "properties": {
            **BASE_MAPPING,
            **GROUP_MAPPING,
            Field.NUMERIC: make_object_type(NUMERIC_MAPPING),
            Field.PROPERTIES: make_object_type(properties),
        },
    }


def make_schema_mapping(schemata: Iterable[SchemaType]) -> Mapping:
    """Create an entity mapping for given schemata with dynamic property resolution."""
    # Multiple schemata can have the same property name, but we flatten them
    # into a single field in the search index with probably multiple copy_to
    # targets. keyword type always has precedence over text.  All fields within
    # a group (text/keyword) are usually the same.  Currently, only "authority"
    # causes a collision (some are string, some are entity)
    merged_props: dict[str, dict[str, set[str]]] = ddict(lambda: ddict(set[str]))

    for schema_name in schemata:
        schema = model.get(schema_name)
        assert schema is not None, schema_name
        for name, prop in schema.properties.items():
            if prop.stub:
                continue
            merged_props[name]["type"].add(get_index_field_type(prop.type))
            if prop.type == registry.text:
                merged_props[name]["copy_to"].add(Field.CONTENT)
            else:
                merged_props[name]["copy_to"].add(Field.TEXT)
            if prop.type.group:
                merged_props[name]["copy_to"].add(prop.type.group)
            if name in schema.caption:
                merged_props[name]["copy_to"].add(Field.NAME)

    # clean up properties type
    properties: dict[str, MappingProperty] = {}
    for prop, config in merged_props.items():
        spec: MappingProperty = {"copy_to": list(config.pop("copy_to"))}
        type_ = config.pop("type")
        if "keyword" in type_:
            type_.discard("text")
        type_ = list(type_)
        assert len(type_) == 1, type_
        properties[prop] = {**spec, "type": type_[0]}

    return properties


def get_index_field_type(type_, to_numeric: bool | None = False) -> str:
    """Given a FtM property type, return the corresponding ElasticSearch field
    type (used for determining the sorting field)"""
    es_type = TYPE_MAPPINGS.get(type_, FieldType.KEYWORD)
    if to_numeric and type_ in NUMERIC_TYPES:
        es_type = FieldType.NUMERIC
    if es_type:
        return es_type.get("type") or FieldType.KEYWORD["type"]
    return FieldType.KEYWORD["type"]


def get_field_type(field) -> str:
    field = field.split(".")[-1]
    if field in registry.groups:
        return str(registry.groups[field])
    for prop in model.properties:
        if prop.name == field:
            return str(prop.type)
    return str(registry.string)



import asyncio
import itertools
from datetime import datetime
from typing import Any, Generator, Iterable, TypeAlias, TypedDict

from anystore.decorators import error_handler
from anystore.io import logged_items
from anystore.logging import get_logger
from elasticsearch import AsyncElasticsearch
from elasticsearch.helpers import BulkIndexError, async_bulk, bulk

from openaleph_search.core import (
    get_async_ingest_es,
    get_es,
    get_ingest_es,
)
from openaleph_search.index.util import (
    check_response,
    check_settings_changed,
    refresh_sync,
)
from openaleph_search.settings import Settings

log = get_logger(__name__)
settings = Settings()

MAX_TIMEOUT = "700m"
MAX_REQUEST_TIMEOUT = 84600


class Action(TypedDict):
    _id: str
    _index: str
    _source: dict[str, Any]


Actions: TypeAlias = Generator[Action, None, None] | Iterable[Action]


@error_handler(logger=log, max_retries=settings.max_retries)
def query_delete(index, query, sync=False, **kwargs):
    "Delete all documents matching the given query inside the index."
    es = get_es()
    return es.delete_by_query(
        index=index,
        body={"query": query},
        # _source=False,
        # slices="auto",
        conflicts="proceed",
        wait_for_completion=sync,
        refresh=refresh_sync(sync),
        timeout=f"{MAX_REQUEST_TIMEOUT}s",
        scroll_size=settings.index_delete_by_query_batchsize,
        **kwargs,
    )


def bulk_actions(
    actions: Actions,
    chunk_size: int | None = settings.indexer_chunk_size,
    max_concurrency: int | None = settings.indexer_concurrency,
    sync: bool | None = False,
):
    """Bulk indexing with parallel async processing - entry point for sync
    applications

    Args:
        actions: Iterator/iterable of actions to index
        chunk_size: Number of actions per chunk
        max_concurrency: Maximum number of concurrent chunks
        sync: Whether to refresh index after operations
    """
    # shortcut for 1 worker
    if max_concurrency == 1:
        es = get_ingest_es()
        return bulk(
            es,
            actions,
            max_retries=settings.max_retries,
            chunk_size=settings.indexer_chunk_size,
            max_chunk_bytes=settings.indexer_max_chunk_bytes,
        )

    return asyncio.run(bulk_actions_async(actions, chunk_size, max_concurrency, sync))


@error_handler(logger=log, max_retries=settings.max_retries)
async def process_chunk(es: AsyncElasticsearch, chunk_actions, sync: bool):
    try:
        result = await async_bulk(
            es,
            chunk_actions,
            max_retries=settings.max_retries,
            refresh=refresh_sync(sync),
            timeout=f"{MAX_REQUEST_TIMEOUT}s",
            request_timeout=MAX_REQUEST_TIMEOUT,  # Client-side timeout
            chunk_size=settings.indexer_chunk_size,
            max_chunk_bytes=settings.indexer_max_chunk_bytes,
        )
        success, failed = result
        for failure in failed:
            if failure.get("delete", {}).get("status") == 404:
                continue
            log.warning("Bulk index error: %r" % failure)
        return success, failed
    except BulkIndexError as e:
        log.error(f"BulkIndexError: {len(e.errors)} document(s) failed to index")
        log.error(f"Error details: {e}")

        # Log detailed information about each failed document
        for i, error in enumerate(e.errors[:10]):  # Log first 10 errors to avoid spam
            log.error(f"Document {i + 1} error: {error}")

        if len(e.errors) > 10:
            log.error(f"... and {len(e.errors) - 10} more errors (truncated)")

        # Re-raise the exception to maintain existing error handling
        raise


async def bulk_actions_async(
    actions: Actions,
    chunk_size: int | None = settings.indexer_chunk_size,
    max_concurrency: int | None = settings.indexer_concurrency,
    sync: bool | None = False,
):
    """Process chunks as they complete to limit memory usage."""
    start = datetime.now()
    es = await get_async_ingest_es()
    actions = logged_items(actions, "Loading", 10_000, item_name="doc", logger=log)
    chunks = itertools.batched(actions, n=chunk_size or settings.indexer_chunk_size)
    max_concurrency = max_concurrency or settings.indexer_concurrency
    semaphore = asyncio.Semaphore(max_concurrency)

    async def process_chunk_with_semaphore(chunk):
        async with semaphore:
            return await process_chunk(es, chunk, sync)

    success = 0
    errors = 0
    pending_tasks = set()

    try:
        for chunk in chunks:
            # Create task
            task = asyncio.create_task(process_chunk_with_semaphore(list(chunk)))
            pending_tasks.add(task)

            # Process completed tasks when we hit concurrency limit
            if len(pending_tasks) >= max_concurrency:
                done, pending_tasks = await asyncio.wait(
                    pending_tasks, return_when=asyncio.FIRST_COMPLETED
                )

                for task in done:
                    try:
                        result = await task
                        success += result[0]
                        errors += len(result[1])
                    except Exception as e:
                        log.error(f"Chunk processing failed: {e}")
                        errors += 1

        # Process remaining tasks
        if pending_tasks:
            results = await asyncio.gather(*pending_tasks, return_exceptions=True)
            for result in results:
                if isinstance(result, Exception):
                    log.error(f"Chunk processing failed: {result}")
                    errors += 1
                else:
                    success += result[0]
                    errors += len(result[1])

    finally:
        await es.close()

    end = datetime.now()
    log.info(
        "Bulk indexing completed: %d successful, %d failed" % (success, errors),
        took=end - start,
    )


@error_handler(logger=log, max_retries=settings.max_retries)
def index_safe(index, id, body, sync=False, **kwargs):
    """Index a single document and retry until it has been stored."""
    es = get_ingest_es()
    refresh = refresh_sync(sync)
    es.index(index=index, id=id, body=body, refresh=refresh, **kwargs)
    body["id"] = str(id)
    body.pop("text", None)
    return body


@error_handler(logger=log, max_retries=settings.max_retries)
def delete_safe(index: str, id: str, sync: bool | None = False):
    es = get_es()
    es.delete(index=index, id=id, ignore=[404], refresh=refresh_sync(sync))


@error_handler(logger=log, max_retries=settings.max_retries)
def rewrite_mapping_safe(pending, existing):
    """This re-writes mappings for ElasticSearch in such a way that
    immutable values are kept to their existing setting, while other
    fields are updated."""
    IMMUTABLE = ("type", "analyzer", "normalizer", "index", "store")
    # This is a pretty bad idea long-term. We need to make it easier
    # to use multiple index generations instead.
    if not isinstance(pending, dict) or not isinstance(existing, dict):
        return pending
    for key, value in list(pending.items()):
        old_value = existing.get(key)
        value = rewrite_mapping_safe(value, old_value)
        if key in IMMUTABLE and old_value is not None:
            value = old_value
        pending[key] = value
    for key, value in existing.items():
        if key not in pending:
            pending[key] = value
    return pending


@error_handler(logger=log, max_retries=settings.max_retries)
def configure_index(index, mapping, settings_):
    """Create or update a search index with the given mapping and
    SETTINGS. This will try to make a new index, or update an
    existing mapping with new properties.
    """
    es = get_es()
    if es.indices.exists(index=index):
        log.info("Configuring index: %s..." % index)
        options = {
            "index": index,
            "timeout": MAX_TIMEOUT,
            "master_timeout": MAX_TIMEOUT,
        }
        config = es.indices.get(index=index).get(index, {})
        settings_.get("index").pop("number_of_shards", settings.index_shards)
        if check_settings_changed(settings_, config.get("settings")):
            res = es.indices.close(ignore_unavailable=True, **options)
            res = es.indices.put_settings(body=settings_, **options)
            if not check_response(index, res):
                return False
        mapping = rewrite_mapping_safe(mapping, config.get("mappings"))
        res = es.indices.put_mapping(body=mapping, **options)
        if not check_response(index, res):
            return False
        res = es.indices.open(**options)
        return True
    else:
        log.info("Creating index: %s..." % index)
        body = {"settings": settings_, "mappings": mapping}
        res = es.indices.create(index=index, body=body)
        if not check_response(index, res):
            return False
        return True



"""
Parse Elasticsearch query string syntax to extract field filters.

Based on Lucene query syntax specification from:
https://lucene.apache.org/core/2_9_4/queryparsersyntax.html
https://github.com/pyparsing/pyparsing/blob/master/examples/lucene_grammar.py
"""

import pyparsing as pp

# Enable packrat parsing for performance
pp.ParserElement.enable_packrat()


def parse_field_filters(
    query_text: str,
) -> tuple[str, dict[str, list[str]], dict[str, list[str]]]:
    """
    Parse field filters from an Elasticsearch query string using Lucene grammar.

    Supports full Lucene/Elasticsearch query syntax including:
    - Field filters: field:value, field:"quoted value"
    - Positive/Negative filters: +field:value, -field:value
    - Range queries: field:[min TO max], field:{min TO max}
    - Fuzzy search: field:value~0.8
    - Proximity search: field:"phrase"~10
    - Wildcards: field:val*, field:te?t
    - Regular expressions: field:/regex/
    - Boosting: field:value^2, field:"phrase"^2.5
    - Comparison operators: field:>10, field:>=10, field:<100, field:<=100
    - Wildcard field names: book.*:value
    - Escaped field names: first\\ name:value
    - Escaped characters: field:"value with \\" quote"
    - Boolean operators: AND, OR, NOT
    - Elasticsearch extensions: _exists_:field, _missing_:field

    Args:
        query_text: The query string to parse

    Returns:
        A tuple of (remaining_text, filters, excludes) where:
        - remaining_text: Query text with field filters removed
        - filters: Dict mapping field names to lists of values
        - excludes: Dict mapping field names to lists of excluded values

    Examples:
        >>> parse_field_filters('name:"Jane Doe" countries:USA')
        ('', {'name': ['Jane Doe'], 'countries': ['USA']}, {})

        >>> parse_field_filters('foo bar schema:Person')
        ('foo bar', {'schema': ['Person']}, {})

        >>> parse_field_filters('test -schema:Page')
        ('test', {}, {'schema': ['Page']})

        >>> parse_field_filters('date:[2020 TO 2022]')
        ('', {'date': ['[ 2020 TO 2022 ]']}, {})

        >>> parse_field_filters('name:john~0.8')
        ('', {'name': ['john~0.8']}, {})

        >>> parse_field_filters('title:"jakarta apache"~10')
        ('', {'title': ['"jakarta apache"~10']}, {})

        >>> parse_field_filters('title:quick^2')
        ('', {'title': ['quick^2']}, {})

        >>> parse_field_filters('age:>10')
        ('', {'age': ['>10']}, {})
    """
    if not query_text or not query_text.strip():
        return "", {}, {}

    filters: dict[str, list[str]] = {}
    excludes: dict[str, list[str]] = {}
    remaining_parts: list[str] = []

    # Define grammar elements
    COLON = pp.Literal(":")
    MINUS = pp.Literal("-")
    PLUS = pp.Literal("+")
    CARAT = pp.Literal("^")

    # Field name: alphanumeric, underscore, dot, hyphen, wildcard, or escaped spaces
    # Based on Lucene spec and Elasticsearch extensions
    # Support: book.author, field_name, book.*, first\ name
    field_name_part = pp.Regex(r'(?:[a-zA-Z0-9_.\-*]|\\ )+')
    field_name = field_name_part("field")

    # Boost modifier: ^[number]
    # Can follow any value type
    boost_modifier = CARAT + pp.Regex(r'[0-9.]+')

    # Range queries: field:[min TO max] or field:{min TO max}
    TILDE = pp.Literal("~")
    LBRACK = pp.Literal("[")
    RBRACK = pp.Literal("]")
    LBRACE = pp.Literal("{")
    RBRACE = pp.Literal("}")
    TO = pp.CaselessKeyword("TO")

    # Range term: anything except TO keyword and brackets
    range_term = pp.Regex(r'[^\s\[\]{}"]+')

    # Inclusive range [min TO max]
    incl_range = LBRACK + range_term + TO + range_term + RBRACK
    # Exclusive range {min TO max}
    excl_range = LBRACE + range_term + TO + range_term + RBRACE

    # Range value with optional boost
    # Custom parsing to format: "[ min TO max ]" or "[ min TO max ]^2"
    def format_range(tokens):
        # tokens[0] is the combined string from Combine, e.g., "[ 2020 TO 2022 ] ^ 2"
        # We need to remove space before ^
        if len(tokens) == 1:
            result = str(tokens[0])
            # Remove spaces around ^
            import re
            result = re.sub(r'\s+\^\s*', '^', result)
            return result
        return str(tokens[0])

    range_value = pp.Combine(
        (incl_range | excl_range) + pp.Optional(boost_modifier),
        adjacent=False,
        join_string=" "
    )("value")
    range_value.set_parse_action(format_range)

    # Quoted string with escape support
    quoted_string_base = pp.QuotedString('"', esc_char="\\")

    # Proximity search: "phrase"~10^2 - must preserve quotes and modifiers
    # Match the entire structure including quotes, proximity, and boost
    quoted_with_proximity = pp.Combine(
        pp.Literal('"')
        + pp.Regex(r'(?:[^"\\]|\\.)*')  # Content with escape support
        + pp.Literal('"')
        + TILDE  # Require tilde
        + pp.Optional(pp.Regex(r'[0-9.]+'))
        + pp.Optional(boost_modifier),  # Optional boost after proximity
        adjacent=True,
        join_string=""
    )

    # Regular quoted string with optional boost - quotes are removed
    quoted_with_boost = pp.Combine(
        quoted_string_base + boost_modifier,
        adjacent=True,
        join_string=""
    )
    quoted_without_boost = quoted_string_base

    # Try proximity version first (most specific), then boost, then plain
    quoted_value = (
        quoted_with_proximity | quoted_with_boost | quoted_without_boost
    )("value")

    # Regular expression value: /regex/
    # Content between forward slashes
    SLASH = pp.Literal("/")
    regex_content = pp.Regex(r'[^/]+')  # Everything between slashes
    regex_value = pp.Combine(
        SLASH + regex_content + SLASH + pp.Optional(boost_modifier),
        adjacent=True,
        join_string=""
    )("value")

    # Unquoted value with optional fuzzy and boost modifiers
    # For fuzzy search: word~0.8^2, word~0.8, word^2
    fuzzy_modifier = TILDE + pp.Optional(pp.Regex(r'[0-9.]+'))
    unquoted_word = pp.Regex(r'[^\s:()[\]{}"\-+~/]+')

    # Try all combinations: fuzzy+boost, fuzzy only, boost only, plain
    unquoted_with_fuzzy_boost = pp.Combine(
        unquoted_word + fuzzy_modifier + boost_modifier,
        adjacent=True,
        join_string=""
    )
    unquoted_with_fuzzy = pp.Combine(
        unquoted_word + fuzzy_modifier,
        adjacent=True,
        join_string=""
    )
    unquoted_with_boost = pp.Combine(
        unquoted_word + boost_modifier,
        adjacent=True,
        join_string=""
    )
    unquoted_plain = unquoted_word

    unquoted_value = (
        unquoted_with_fuzzy_boost
        | unquoted_with_fuzzy
        | unquoted_with_boost
        | unquoted_plain
    )("value")

    # Value can be range, quoted, regex, or unquoted (order matters)
    value = range_value | quoted_value | regex_value | unquoted_value

    # Optional positive/negative modifiers
    # + means required (we treat as normal filter)
    # - means excluded (negation)
    modifier = pp.Optional(MINUS | PLUS)("modifier")

    # Field filter pattern: [+/-]field:value
    field_filter = pp.Group(modifier + field_name + COLON + value)

    # Scan for all field filter matches in the query
    last_end = 0
    for tokens, start, end in field_filter.scan_string(query_text):
        # Add any text before this match to remaining parts
        if start > last_end:
            text_before = query_text[last_end:start].strip()
            if text_before:
                # Filter out standalone operators
                words = text_before.split()
                for word in words:
                    if word.upper() not in ("AND", "OR", "NOT"):
                        remaining_parts.append(word)

        # Extract the matched filter
        match = tokens[0]  # type: ignore[index]
        mod = str(match.get("modifier", ""))  # type: ignore[union-attr]
        is_negated = mod == "-"
        field = str(match["field"])  # type: ignore[index]
        val = str(match["value"])  # type: ignore[index]

        # Clean up escaped spaces in field names
        field = field.replace("\\ ", " ")

        # Add to appropriate dict
        target = excludes if is_negated else filters
        if field not in target:
            target[field] = []
        target[field].append(val)

        last_end = end

    # Add any remaining text after the last match
    if last_end < len(query_text):
        text_after = query_text[last_end:].strip()
        if text_after:
            words = text_after.split()
            for word in words:
                if word.upper() not in ("AND", "OR", "NOT"):
                    remaining_parts.append(word)

    remaining_text = " ".join(remaining_parts)
    return remaining_text, filters, excludes



"""Transform followthemoney.EntityProxy into index actions"""

import functools
import itertools
import time
from concurrent.futures import ProcessPoolExecutor
from datetime import datetime
from multiprocessing import cpu_count
from typing import Generator, Iterable

from anystore.functools import weakref_cache
from anystore.logging import get_logger
from anystore.util import Took
from banal import ensure_list
from followthemoney import EntityProxy, model, registry
from followthemoney.namespace import Namespace

from openaleph_search.index.indexer import Action, Actions
from openaleph_search.index.indexes import entities_write_index, schema_bucket
from openaleph_search.index.mapping import NUMERIC_TYPES, Field
from openaleph_search.settings import Settings, __version__
from openaleph_search.transform.tagging import (
    get_name_symbols,
    get_symbols,
    select_symbols,
)
from openaleph_search.transform.util import (
    get_geopoints,
    index_name_keys,
    index_name_parts,
    phonetic_names,
)
from openaleph_search.util import valid_dataset

log = get_logger(__name__)
settings = Settings()


def _numeric_values(type_, values) -> list[float]:
    values = [type_.to_number(v) for v in ensure_list(values)]
    return [v for v in values if v is not None]


def _get_symbols(entity: EntityProxy) -> set[str]:
    symbols = select_symbols(entity)  # pre-computed in earlier stage
    if symbols:
        return symbols
    if entity.schema.is_a("LegalEntity"):
        return {str(s) for s in get_symbols(entity)}
    symbols: set[str] = set()
    symbols.update(map(str, get_name_symbols(model["Person"], *entity.names)))
    symbols.update(map(str, get_name_symbols(model["Organization"], *entity.names)))
    return symbols


@weakref_cache
def _get_namespace(value: str) -> Namespace:
    return Namespace(value)


def format_entity(dataset: str, entity: EntityProxy, **kwargs) -> Action | None:
    """Apply final denormalisations to the index."""
    # Abstract entities can appear when profile fragments for a missing entity
    # are present.
    if entity.schema.abstract:
        log.warning(
            "Tried to index an abstract-typed entity!",
            schema=entity.schema.name,
            entity_id=entity.id,
        )
        return None

    if settings.index_namespace_ids:
        # Enforce namespaced IDs
        ns = _get_namespace(dataset)
        entity = ns.apply(entity)

    dataset = valid_dataset(dataset)

    data = entity.to_dict()
    # deprecated
    collection_id = kwargs.get("collection_id")
    if collection_id is not None:
        data[Field.COLLECTION_ID] = collection_id

    data[Field.DATASET] = dataset
    data[Field.SCHEMATA] = list(entity.schema.names)
    data[Field.CAPTION] = entity.caption

    # all names, including mentioned ones, for lookups
    names = list(entity.names)
    symbols = list(_get_symbols(entity))
    if symbols:
        data[Field.NAME_SYMBOLS] = symbols
    name_keys = list(index_name_keys(entity.schema, names))
    if name_keys:
        data[Field.NAME_KEYS] = name_keys
    name_parts = list(index_name_parts(entity.schema, names))
    if name_parts:
        data[Field.NAME_PARTS] = name_parts
    name_phonetics = list(phonetic_names(entity.schema, names))
    if name_phonetics:
        data[Field.NAME_PHONETIC] = name_phonetics

    # Add tags from EntityProxy.context (they are added from aleph db before indexing)
    tags = ensure_list(entity.context.get("tags"))
    if tags:
        data[Field.TAGS] = tags

    # Slight hack: a magic property in followthemoney that gets taken out
    # of the properties and added straight to the index text.
    properties = data.get("properties", {})
    text = properties.pop("indexText", [])
    if text:
        data[Field.CONTENT] = text

    # length normalization
    data[Field.NUM_VALUES] = sum([len(v) for v in properties.values()])

    # integer casting
    numeric = {}
    for prop in entity.iterprops():
        if prop.type in NUMERIC_TYPES:
            values = entity.get(prop)
            numeric[prop.name] = _numeric_values(prop.type, values)
    # also cast group field for dates
    dates = _numeric_values(registry.date, entity.get_type_values(registry.date))
    if dates:
        numeric["dates"] = dates
    if numeric:
        data[Field.NUMERIC] = numeric

    # geo data if entity is an Address
    if "latitude" in entity.schema.properties:
        data[Field.GEO_POINT] = get_geopoints(entity)

    # Context data - from aleph system, not followthemoney. Probably deprecated soon
    if hasattr(entity, "context"):
        role_id = entity.context.get("role_id")
        if role_id:
            data[Field.ROLE] = role_id
        profile_id = entity.context.get("profile_id")
        if profile_id:
            data[Field.PROFILE] = profile_id
        origin = ensure_list(entity.context.get("origin"))
        if origin:
            data[Field.ORIGIN] = origin
        data[Field.MUTABLE] = entity.context.get("mutable", False) or False
        # Logical simplifications of dates:
        created_at = ensure_list(entity.context.get("created_at"))
        if len(created_at) > 0:
            data[Field.CREATED_AT] = min(created_at)
        updated_at = ensure_list(entity.context.get("updated_at")) or created_at
        if len(updated_at) > 0:
            data[Field.UPDATED_AT] = max(updated_at)

    data[Field.INDEX_BUCKET] = schema_bucket(data["schema"])
    data[Field.INDEX_VERSION] = __version__
    data[Field.INDEX_TS] = datetime.now().isoformat()

    # log.info("%s", pformat(data))
    entity_id = data.pop("id")
    return {
        "_id": entity_id,
        "_index": entities_write_index(entity.schema),
        "_source": data,
    }


def format_entities(dataset: str, entities: Iterable[EntityProxy], **kwargs) -> Actions:
    for entity in entities:
        formatted = format_entity(dataset, entity, **kwargs)
        if formatted is not None:
            yield formatted


def format_batch(
    dataset: str, entities: Iterable[EntityProxy], **kwargs
) -> list[Action]:
    actions = []
    for entity in entities:
        formatted = format_entity(dataset, entity, **kwargs)
        if formatted is not None:
            actions.append(formatted)
    return actions


def format_parallel(
    dataset: str,
    entities: Generator[EntityProxy, None, None],
    concurrency: int | None = settings.indexer_concurrency,
    chunk_size: int | None = settings.indexer_chunk_size,
    **kwargs,
) -> Actions:
    """
    Transform entities into index actions in parallel
    """
    batches = itertools.batched(entities, n=chunk_size or settings.indexer_chunk_size)
    max_workers = min((cpu_count(), concurrency or settings.indexer_concurrency))
    max_queued = max_workers * 2
    func = functools.partial(format_batch, dataset=dataset, **kwargs)

    # shortcut for 1 worker
    if max_workers == 1:
        yield from func(entities=entities)
        return

    with ProcessPoolExecutor(max_workers=max_workers) as executor, Took() as t:
        transformed = 0
        active_futures = {}
        # Submit initial batches
        for _ in range(max_queued):
            try:
                batch = next(batches)
                future = executor.submit(func, entities=batch)
                active_futures[hash(future)] = future
            except StopIteration:
                break

        # Process results as they complete
        while active_futures:
            # Wait for at least one to complete
            completed = [f for f in active_futures.values() if f.done()]
            for future in completed:
                for action in future.result():
                    transformed += 1
                    yield action
                del active_futures[hash(future)]

                # Submit next batch
                try:
                    batch = next(batches)
                    new_future = executor.submit(func, entities=batch)
                    active_futures[hash(new_future)] = new_future
                except StopIteration:
                    pass

            if not completed:
                # If none completed, wait a bit
                time.sleep(0.1)

    log.info(f"Transformed {transformed} actions.", took=t.took)
